from .shape import Shape

class Ellipse(Shape):
    def __init__(self, cx, cy, rx, ry, **kwargs):
        super().__init__(**kwargs)
        self.tag = "ellipse"
        self.attributes['cx'] = cx
        self.attributes['cy'] = cy
        self.attributes['rx'] = rx
        self.attributes['ry'] = ry


from .element import Element

class Text(Element):
    def __init__(self, x, y, content, **kwargs):
        super().__init__(**kwargs)
        self.tag = "text"
        self.attributes['x'] = x
        self.attributes['y'] = y
        self.content = content
        self.styles['font-family'] = kwargs.get('font', None)
        self.styles['font-size'] = kwargs.get('size', None)
        self.styles['text-anchor'] = kwargs.get('anchor', None)
        self.styles['dominant-baseline'] = kwargs.get('baseline', None)


from savage import Graphic, Circle, Rect

def test_graphic_creation():
    graphic = Graphic(width=500, height=400)

    assert graphic.width == 500
    assert graphic.height == 400
    assert graphic.elements == []

def test_graphic_add_elements():
    graphic = Graphic(width=500, height=400)
    circle = Circle(cx=50, cy=50, r=20)
    rect = Rect(width=100, height=50, x=150, y=150)

    graphic.add(circle)
    graphic.add(rect)

    assert len(graphic.elements) == 2
    assert graphic.elements[0] is circle
    assert graphic.elements[1] is rect

def test_graphic_to_svg():
    graphic = Graphic(width=500, height=400)
    circle = Circle(cx=50, cy=50, r=20, fill="red")
    rect = Rect(width=100, height=50, x=150, y=150, fill="blue")

    graphic.add(circle)
    graphic.add(rect)

    svg_output = graphic.to_svg()

    # Check that the SVG output contains the correct XML structure
    assert '<?xml version="1.0" encoding="UTF-8" standalone="no"?>' in svg_output
    assert '<svg xmlns="http://www.w3.org/2000/svg" width="500" height="400">' in svg_output
    assert '<circle' in svg_output  # Verify Circle element is included
    assert 'fill: red;' in svg_output  # Verify the fill of the circle
    assert '<rect' in svg_output  # Verify Rect element is included
    assert 'fill: blue;' in svg_output  # Verify the fill of the rect
    assert svg_output.endswith('</svg>')  # Ensure the SVG tag closes correctly

def test_graphic_save(mocker):
    # Mock the open function to avoid actually writing to the file system
    mock_open = mocker.patch("builtins.open", mocker.mock_open())
    
    graphic = Graphic(width=500, height=400)
    circle = Circle(cx=50, cy=50, r=20)
    graphic.add(circle)

    graphic.save("test_output.svg")

    # Check if open was called with the correct filepath
    mock_open.assert_called_once_with("test_output.svg", 'w')
    
    # Check that the file contents match the expected SVG output
    svg_output = graphic.to_svg()
    mock_open.return_value.write.assert_called_once_with(svg_output)

def test_svg_element_order_preserved():
    graphic = Graphic(width=500, height=400)
    rect = Rect(width=100, height=50, x=10, y=10)
    circle = Circle(cx=20, cy=20, r=10)

    graphic.add(rect)
    graphic.add(circle)

    svg_output = graphic.to_svg()

    # rect should appear before circle
    rect_index = svg_output.index("<rect")
    circle_index = svg_output.index("<circle")
    assert rect_index < circle_index



from .element import Element
from .group import Group
from .text import Text
from .shapes import Shape
from .shapes import Circle
from .shapes import Ellipse
from .shapes import Line
from .shapes import Polygon
from .shapes import Polyline
from .shapes import Rect


__all__ = ["Element", "Group", "Text", "Shape", "Circle", "Ellipse", "Line", "Polygon", "Polyline", "Rect"]


from .shape import Shape

class Circle(Shape):
    def __init__(self, cx, cy, r, **kwargs):
        super().__init__(**kwargs)
        self.tag = "circle"
        self.attributes['cx'] = cx
        self.attributes['cy'] = cy
        self.attributes['r'] = r


from .shape import Shape

class Rect(Shape):
    def __init__(self, width, height, x, y, rx=None, ry=None, **kwargs):
        super().__init__(**kwargs)
        self.tag = "rect"
        self.attributes['width'] = width
        self.attributes['height'] = height
        self.attributes['x'] = x
        self.attributes['y'] = y
        if rx:
            self.attributes['rx'] = rx
        if ry:
            self.attributes['ry'] = ry


from .shape import Shape

class Line(Shape):
    def __init__(self, x1, y1, x2, y2, **kwargs):
        super().__init__(**kwargs)
        self.tag = "line"
        self.attributes['x1'] = x1
        self.attributes['y1'] = y1
        self.attributes['x2'] = x2
        self.attributes['y2'] = y2


from ..element import Element

class Shape(Element):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def to_svg(self, indentation=2):
        svg_string = indentation * " "
        svg_string += self.add_open_tag_svg()
        svg_string = svg_string.replace(">"," />")
        return svg_string


from .elements import Element, Shape

class Graphic:
    def __init__(self, width=None, height=None, background=None, **kwargs):
        self.width = width
        self.height = height
        self.background = background
        self.elements: list[Element] = []
        self.tags_used = []
        self.text_used = False
        self.default_styles = {}
        self.default_styles['fill'] = kwargs.get('fill', '#EBF3F7')
        self.default_styles['stroke'] = kwargs.get('stroke', '#9CA7AD')
        self.default_styles['stroke-width'] = kwargs.get('strokewidth', 2)
        
    def add(self, element:Element):
        self.elements.append(element)
        tag = element.tag
        # if isinstance(element, Shape) and element.tag not in self.tags_used:
        ignored_elements = ['g', 'text']
        if element.tag not in self.tags_used and element.tag not in ignored_elements:
            self.tags_used.append(element.tag)
            self.tags_used.sort()
        if element.tag == 'text':
            self.text_used = True

    def to_svg(self):
        # TODO: allow default styles to be specified for shapes
        svg_string = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n"""
        width_svg = f' width="{self.width}"' if self.width is not None else ""
        height_svg = f' height="{self.height}"' if self.height is not None else ""
        svg_string += f"""<svg xmlns="http://www.w3.org/2000/svg"{width_svg}{height_svg}>\n"""
        if self.tags_used or self.text_used:
            svg_string += "  <style>\n"
        if self.tags_used:
            svg_string += f'''    {", ".join(self.tags_used)} {{\n'''
            svg_string += f'''      fill: {self.default_styles['fill']};\n'''
            svg_string += f'''      stroke: {self.default_styles['stroke']};\n'''
            svg_string += f'''      stroke-width: {self.default_styles['stroke-width']};\n'''
            svg_string += '    }\n'
        if self.text_used:
            svg_string += "    text { font-family: sans-serif; text-anchor: middle; dominant-baseline: middle; fill: #40484D; }\n"
        if self.tags_used or self.text_used:
            svg_string += "  </style>\n"
        if self.background:
            svg_string += f'  <rect width="100%" height="100%" style="fill: {self.background}; stroke: none" />\n'
        for element in self.elements:
            # svg_string += "  " + element.to_svg() + "\n"
            svg_string += element.to_svg() + "\n"
        svg_string += """</svg>"""
        return svg_string
    
    def save(self, filepath):
        svg_string = self.to_svg()
        with open(filepath, 'w') as output:
            output.write(svg_string)


from .shape import Shape
from .circle import Circle
from .ellipse import Ellipse
from .line import Line
from .polygon import Polygon
from .polyline import Polyline
from .rect import Rect

__all__ = ["Shape", "Circle", "Ellipse", "Line", "Polygon", "Polyline", "Rect"]


from .shape import Shape

class Polygon(Shape):
    def __init__(self, points:list[int], **kwargs):
        super().__init__(**kwargs)
        self.tag = "polygon"
        if len(points) % 2 != 0:
            raise ValueError("Points must contain an even number of elements (x, y pairs).")
        self.attributes['points'] = points
        
    def add_attribute_svg(self):
        points_svg = " ".join([f"{self.attributes['points'][i]},{self.attributes['points'][i+1]}" for i in range(0, len(self.attributes['points']), 2)])
        return f' points="{points_svg}"'


class Element:
    def __init__(self, **kwargs):
        self.tag = None # defined in child classes
        self.attributes = {}
        self.content = None # defined in child classes where relevant
        style_attributes = {'fill':'fill', 'stroke':'stroke', 'strokewidth':'stroke-width', 'opacity':'opacity'}
        self.styles = {}
        for style_argument, style_attribute in style_attributes.items():
            self.styles[style_attribute] = kwargs.get(style_argument, None)
        self.transformations = []
        
    def translate(self, dx, dy):
        self.transformations.append(f"translate({dx},{dy})")
        return self
    
    def rotate(self, angle, cx=None, cy=None):
        if cx is not None and cy is not None:
            self.transformations.append(f"rotate({angle} {cx} {cy})")
        else:
            self.transformations.append(f"rotate({angle})")
        return self
    
    def scale(self, sx, sy=None):
        if sy is not None:
            self.transformations.append(f"scale({sx} {sy})")
        else:
            self.transformations.append(f"scale({sx})")
        return self
    
    def skew(self, x=None, y=None):
        if x is not None:
            self.transformations.append(f"skewX({x})")
        if y is not None:
            self.transformations.append(f"skewY({y})")
        return self

    def to_svg(self, indentation=2):
        svg_string = indentation * " "
        svg_string += self.add_open_tag_svg()
        svg_string += self.add_content_svg(indentation)
        svg_string += f"</{self.tag}>"
        return svg_string

    def add_open_tag_svg(self):
        open_tag_svg = f"<{self.tag}{self.add_attribute_svg()}{self.add_transform_svg()}{self.add_style_svg()}>"
        return open_tag_svg

    def add_attribute_svg(self):
        attribute_svg = ""
        for attribute,value in self.attributes.items():
            attribute_svg += f' {attribute}="{value}"'
        return attribute_svg
    
    def add_style_svg(self):
        styling_svg = ""
        inline_styles = []
        for style_attribute, value in self.styles.items():
            if value is not None:
                inline_styles.append(f"{style_attribute}: {value};")
        if inline_styles:
            styling_svg = f' style="{" ".join(inline_styles)}"'
        return styling_svg

    def add_transform_svg(self):
        if self.transformations:
            return f''' transform="{" ".join(self.transformations)}"'''
        return ""

    def add_content_svg(self, indentation=None):
        return self.content


from .element import Element

class Group(Element):
    def __init__(self, content=None, **kwargs):
        super().__init__(**kwargs)
        self.tag = "g"
        self.content: list[Element] = content if content is not None else []
        
    def add(self, element:Element):
        self.content.append(element)
        
    def add_open_tag_svg(self):
        return super().add_open_tag_svg() +"\n"
    
    def add_content_svg(self, indentation=2):
        content_svg = ""
        for element in self.content:
            content_svg += element.to_svg(indentation+2) + "\n"
        return content_svg


from .graphic import Graphic
from .elements import Element, Group, Text, Shape, Circle, Ellipse, Line, Polygon, Polyline, Rect

__all__ = ["Graphic", "Element", "Group", "Text", "Shape", "Circle", "Ellipse", "Line", "Polygon", "Polyline", "Rect"]